package com.qmth.distributed.print.api.obe;

import cn.hutool.core.date.DateUtil;
import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.deepoove.poi.XWPFTemplate;
import com.google.gson.reflect.TypeToken;
import com.qmth.boot.api.annotation.Aac;
import com.qmth.boot.api.constant.ApiConstant;
import com.qmth.boot.api.exception.ApiException;
import com.qmth.boot.core.rateLimit.annotation.RateLimit;
import com.qmth.distributed.print.business.service.PrintCommonService;
import com.qmth.distributed.print.business.util.PdfUtil;
import com.qmth.distributed.print.util.WordToPdfUtil;
import com.qmth.teachcloud.common.annotation.OperationLogDetail;
import com.qmth.teachcloud.common.contant.SystemConstant;
import com.qmth.teachcloud.common.entity.BasicCourse;
import com.qmth.teachcloud.common.entity.BasicSchool;
import com.qmth.teachcloud.common.entity.SysConfig;
import com.qmth.teachcloud.common.entity.SysUser;
import com.qmth.teachcloud.common.enums.ExceptionResultEnum;
import com.qmth.teachcloud.common.enums.FieldUniqueEnum;
import com.qmth.teachcloud.common.enums.ImportTemplateEnum;
import com.qmth.teachcloud.common.enums.log.OperationTypeEnum;
import com.qmth.teachcloud.common.service.BasicCourseService;
import com.qmth.teachcloud.common.service.CommonCacheService;
import com.qmth.teachcloud.common.service.FileUploadService;
import com.qmth.teachcloud.common.util.*;
import com.qmth.teachcloud.mark.entity.MarkPaper;
import com.qmth.teachcloud.mark.enums.LockType;
import com.qmth.teachcloud.mark.lock.impl.CustomRedisLockProvider;
import com.qmth.teachcloud.obe.been.dto.CourseWeightDetailDto;
import com.qmth.teachcloud.obe.been.dto.CourseWeightDto;
import com.qmth.teachcloud.obe.been.dto.TCUsualScoreDto;
import com.qmth.teachcloud.obe.been.report.*;
import com.qmth.teachcloud.obe.been.result.ObeCourseWeightResult;
import com.qmth.teachcloud.obe.been.result.ObeScoreResult;
import com.qmth.teachcloud.obe.been.result.report.PaperStructDimensionResult;
import com.qmth.teachcloud.obe.been.result.report.ReportChangeResult;
import com.qmth.teachcloud.obe.been.result.report.ReportResult;
import com.qmth.teachcloud.obe.been.result.report.word.CourseBasicBean;
import com.qmth.teachcloud.obe.been.result.report.word.CourseReportBean;
import com.qmth.teachcloud.obe.entity.*;
import com.qmth.teachcloud.obe.service.*;
import io.swagger.annotations.*;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.IOUtils;
import org.apache.poi.ss.usermodel.*;
import org.apache.poi.ss.util.CellRangeAddress;
import org.apache.poi.ss.util.RegionUtil;
import org.apache.poi.xssf.streaming.SXSSFWorkbook;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.dao.DuplicateKeyException;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.validation.BindingResult;
import org.springframework.web.bind.annotation.*;

import javax.annotation.Resource;
import javax.validation.Valid;
import javax.validation.constraints.Max;
import javax.validation.constraints.Min;
import java.io.*;
import java.math.BigDecimal;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.text.MessageFormat;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

/**
 * <p>
 * 报告基本情况表 前端控制器
 * </p>
 *
 * @author wangliang
 * @since 2024-02-18
 */
@Api(tags = "课程目标达成度-报告基本情况Controller")
@RestController
@RequestMapping(ApiConstant.DEFAULT_URI_PREFIX + SystemConstant.PREFIX_URL_COURSE_DEGREE)
public class TRBasicInfoController {
    private final static Logger log = LoggerFactory.getLogger(TRBasicInfoController.class);

    @Resource
    PrintCommonService printCommonService;

    @Resource
    TRBasicInfoService trBasicInfoService;

    @Resource
    TRExamStudentService trExamStudentService;

    @Resource
    CommonCacheService commonCacheService;

    @Resource
    FileUploadService fileUploadService;

    @Resource
    TCPaperStructService tcPaperStructService;

    @Resource
    TCUsualScoreService tcUsualScoreService;

    @Resource
    ObeCourseOutlineService obeCourseOutlineService;

    @Resource
    BasicCourseService basicCourseService;

    @Resource
    TCFinalScoreService tcFinalScoreService;

    @Resource
    RedisUtil redisUtil;

    @ApiOperation(value = "报告管理列表")
    @RequestMapping(value = "/report/list", method = RequestMethod.POST)
    @ApiResponses({@ApiResponse(code = 200, message = "分页查询", response = ObeScoreResult.class)})
    public Result reportList(@ApiParam(value = "培养方案ID") @RequestParam(required = false) Long cultureProgramId,
                             @ApiParam(value = "学期ID") @RequestParam(required = false) Long semesterId,
                             @ApiParam(value = "课程id", required = true) Long courseId,
                             @ApiParam(value = "分页页码", required = true) @RequestParam @Min(SystemConstant.PAGE_NUMBER_MIN) Integer pageNumber,
                             @ApiParam(value = "分页数", required = true) @RequestParam @Min(SystemConstant.PAGE_SIZE_MIN) @Max(SystemConstant.PAGE_SIZE_MAX) Integer pageSize) {
        return ResultUtil.ok(
                tcFinalScoreService.scoreList(new Page<>(pageNumber, pageSize), cultureProgramId, semesterId, courseId));
    }

    @ApiOperation(value = "查看报告")
    @RequestMapping(value = "/report/view", method = RequestMethod.POST)
    @ApiResponses({@ApiResponse(code = 200, message = "查看报告", response = ReportResult.class)})
    @Transactional
    public Result reportView(@ApiParam(value = "培养方案id", required = true) @RequestParam Long cultureProgramId,
                             @ApiParam(value = "课程id", required = true) @RequestParam Long courseId,
                             @ApiParam(value = "试卷编号") @RequestParam(required = false) String paperNumber) {
        SysUser sysUser = (SysUser) ServletUtil.getRequestUser();
        BasicCourse basicCourse = basicCourseService.getById(courseId);

        TRBasicInfo trBasicInfo = trBasicInfoService.queryBasicInfo(cultureProgramId, courseId, paperNumber);
        ObeCourseOutline obeCourseOutline = obeCourseOutlineService.findByCultureProgramIdAndCourseId(cultureProgramId,
                courseId);

        ObeCourseWeightResult obeCourseWeightResult = trBasicInfoService.findCourseWeightResultRmi(obeCourseOutline.getId());
        log.info("ObeCourseWeightResult:{}", JacksonUtil.parseJson(obeCourseWeightResult));
        TCPaperStruct tcPaperStruct = tcPaperStructService.queryPaperStruct(cultureProgramId, courseId, paperNumber);
        if (Objects.isNull(tcPaperStruct) || Objects.isNull(tcPaperStruct.getPaperStructDimension())) {
            throw ExceptionResultEnum.ERROR.exception("未找到试卷蓝图信息");
        }
        MarkPaper markPaper = printCommonService.getMarkPaper(null, tcPaperStruct.getPaperNumber(), courseId);
        if (Objects.isNull(trBasicInfo)) {
            trBasicInfo = trBasicInfoService.getReportView(trBasicInfo, markPaper, sysUser.getId(),
                    obeCourseWeightResult, cultureProgramId, courseId);
        } else {
            ReportCourseEvaluationResultDto reportCourseEvaluationResultDto = Objects.nonNull(trBasicInfo.getCourseEvaluationResult()) ?
                    JSONObject.parseObject(trBasicInfo.getCourseEvaluationResult(), ReportCourseEvaluationResultDto.class) :
                    null;
            if (Objects.nonNull(reportCourseEvaluationResultDto)
                    && Objects.nonNull(reportCourseEvaluationResultDto.getWeightSettingSign())
                    && Objects.nonNull(obeCourseWeightResult)
                    && Objects.nonNull(obeCourseWeightResult.getWeightSettingSign())
                    && reportCourseEvaluationResultDto.getWeightSettingSign().longValue() != obeCourseWeightResult.getWeightSettingSign().longValue()) {
                trBasicInfoService.clearReportData(cultureProgramId, courseId, paperNumber,
                        Objects.nonNull(tcPaperStruct.getDimensionSign())
                                && tcPaperStruct.getDimensionSign().longValue()
                                != obeCourseWeightResult.getDimensionSign().longValue());
                trBasicInfo = trBasicInfoService.getReportView(trBasicInfo, markPaper, sysUser.getId(),
                        obeCourseWeightResult, cultureProgramId, courseId);
            } else {
                ReportCourseBasicInfoDto reportCourseBasicInfoDto = new ReportCourseBasicInfoDto(trBasicInfo);
                ReportCourseEvaluationSpreadDto reportCourseEvaluationSpreadDto = Objects.nonNull(trBasicInfo.getCourseEvaluationSpread()) ?
                        JSONObject.parseObject(trBasicInfo.getCourseEvaluationSpread(), ReportCourseEvaluationSpreadDto.class) :
                        null;
                ReportCourseEvaluationResultDetailDto reportCourseEvaluationResultDetailDto = null;
                List<TRExamStudent> trExamStudentList = trExamStudentService.list(
                        new QueryWrapper<TRExamStudent>().lambda().eq(TRExamStudent::getrBasicInfoId, trBasicInfo.getId()));
                if (!CollectionUtils.isEmpty(trExamStudentList)) {
                    List<ReportExamStudentDto> examStudentList = new ArrayList<>(trExamStudentList.size());
                    for (TRExamStudent trExamStudent : trExamStudentList) {
                        examStudentList.add(new ReportExamStudentDto(trExamStudent));
                    }
                    reportCourseEvaluationResultDetailDto = new ReportCourseEvaluationResultDetailDto(examStudentList);
                }
                trBasicInfo.setReportResult(new ReportResult(new ReportCommonDto(null, basicCourse.getCode(), markPaper.getCourseName(), paperNumber),
                        reportCourseBasicInfoDto, reportCourseEvaluationSpreadDto, reportCourseEvaluationResultDto,
                        reportCourseEvaluationResultDetailDto, trBasicInfo));
            }
        }
        trBasicInfo.updateInfo(sysUser.getId());
        //课程目标达成评价明细结果-课程目标达成评价值图
        trBasicInfoService.saveOrUpdate(trBasicInfo);
        return ResultUtil.ok(trBasicInfo.getReportResult());
    }

    @ApiOperation(value = "报告重新计算")
    @RequestMapping(value = "/report/again/calculate", method = RequestMethod.POST)
    @ApiResponses({@ApiResponse(code = 200, message = "查看报告", response = ReportResult.class)})
    @Transactional
    @Aac(rateLimit = @RateLimit(count = 1, period = 1000L))
    public Result reportAgainCalculate(@ApiParam(value = "培养方案id", required = true) @RequestParam Long cultureProgramId,
                                       @ApiParam(value = "课程id", required = true) @RequestParam Long courseId,
                                       @ApiParam(value = "试卷编号") @RequestParam(required = false) String paperNumber) {
        String lockKey = SystemConstant.REDIS_COURSE_DEGREE_DATA_PREFIX + cultureProgramId + "_" + courseId;
        boolean lock = redisUtil.lock(lockKey, SystemConstant.REDIS_COURSE_DEGREE_DATA_TIME_OUT);
        if (!lock) {
            throw ExceptionResultEnum.ERROR.exception("正在重新计算数据，请稍候再试！");
        }
        TRBasicInfo trBasicInfoAgainCalculate = null;
        try {
            SysUser sysUser = (SysUser) ServletUtil.getRequestUser();
            ObeCourseOutline obeCourseOutline = obeCourseOutlineService.findByCultureProgramIdAndCourseId(cultureProgramId,
                    courseId);

            ObeCourseWeightResult obeCourseWeightResult = trBasicInfoService.findCourseWeightResultRmi(obeCourseOutline.getId());
            log.info("ObeCourseWeightResult:{}", JacksonUtil.parseJson(obeCourseWeightResult));
            TCPaperStruct tcPaperStruct = tcPaperStructService.queryPaperStruct(cultureProgramId, courseId, paperNumber);
            if (Objects.isNull(tcPaperStruct) || Objects.isNull(tcPaperStruct.getPaperStructDimension())) {
                throw ExceptionResultEnum.ERROR.exception("未找到试卷蓝图信息");
            }
            MarkPaper markPaper = printCommonService.getMarkPaper(null, tcPaperStruct.getPaperNumber(), courseId);

            TRBasicInfo trBasicInfo = trBasicInfoService.queryBasicInfo(cultureProgramId, courseId, paperNumber);
            trBasicInfoService.clearReportData(cultureProgramId, courseId, paperNumber,
                    Objects.nonNull(tcPaperStruct.getDimensionSign())
                            && tcPaperStruct.getDimensionSign().longValue()
                            != obeCourseWeightResult.getDimensionSign().longValue());

            trBasicInfoAgainCalculate = trBasicInfoService.getReportView(null, markPaper, sysUser.getId(),
                    obeCourseWeightResult, cultureProgramId, courseId);
            if (Objects.nonNull(trBasicInfo)) {
                trBasicInfoAgainCalculate.setCalculate(trBasicInfo);
            }
            //课程目标达成评价明细结果-课程目标达成评价值图
            trBasicInfoService.save(trBasicInfoAgainCalculate);
        } catch (Exception e) {
            log.error(SystemConstant.LOG_ERROR, e);
            if (e instanceof ApiException) {
                ResultUtil.error((ApiException) e, ((ApiException) e).getCode(), e.getMessage());
            } else {
                ResultUtil.error(e.getMessage());
            }
        } finally {
            redisUtil.releaseLock(lockKey);
        }
        return ResultUtil.ok(Objects.nonNull(trBasicInfoAgainCalculate) ? trBasicInfoAgainCalculate.getReportResult() : trBasicInfoAgainCalculate);
    }

    @ApiOperation(value = "保存报告")
    @RequestMapping(value = "/report/save", method = RequestMethod.POST)
    @OperationLogDetail(operationType = OperationTypeEnum.UPDATE)
    @ApiResponses({@ApiResponse(code = 200, message = "保存报告", response = Object.class)})
    @Transactional
    public Result reportSave(@ApiParam(value = "保存报告结构", required = true) @Valid @RequestBody TRBasicInfo trBasicInfo, BindingResult bindingResult) {
        if (bindingResult.hasErrors()) {
            return ResultUtil.error(bindingResult.getAllErrors().get(0).getDefaultMessage());
        }
        SysUser sysUser = (SysUser) ServletUtil.getRequestUser();
        TRBasicInfo trBasicInfoDb = null;
        try {
            trBasicInfoDb = trBasicInfoService.queryBasicInfo(trBasicInfo.getCultureProgramId(), trBasicInfo.getCourseId(), trBasicInfo.getPaperNumber());
            ObeCourseOutline obeCourseOutline = obeCourseOutlineService.findByCultureProgramIdAndCourseId(trBasicInfo.getCultureProgramId(), trBasicInfo.getCourseId());
            ObeCourseWeightResult obeCourseWeightResult = trBasicInfoService.findCourseWeightResultRmi(obeCourseOutline.getId());
            TCPaperStruct tcPaperStruct = tcPaperStructService.queryPaperStruct(trBasicInfo.getCultureProgramId(), trBasicInfo.getCourseId(), trBasicInfo.getPaperNumber());
            MarkPaper markPaper = printCommonService.getMarkPaper(trBasicInfo.getExamId(), Objects.nonNull(tcPaperStruct) ? tcPaperStruct.getPaperNumber() : trBasicInfo.getPaperNumber(), trBasicInfo.getCourseId());
            if (Objects.isNull(trBasicInfoDb)) {
                trBasicInfoDb = new TRBasicInfo(trBasicInfo, sysUser.getId());
                trBasicInfoDb = trBasicInfoService.getReportView(trBasicInfoDb, markPaper, sysUser.getId(), obeCourseWeightResult, trBasicInfo.getCultureProgramId(), trBasicInfo.getCourseId());
            } else {
                ReportCourseBasicInfoDto reportCourseBasicInfoDto = trBasicInfoService.setBasicInfo(trBasicInfo, obeCourseOutline, markPaper, obeCourseWeightResult);
                trBasicInfoDb.updateInfo(trBasicInfo, reportCourseBasicInfoDto, sysUser.getId());
            }
        } catch (Exception e) {
            log.error(SystemConstant.LOG_ERROR, e);
            if (e instanceof DuplicateKeyException) {
                String errorColumn = e.getCause().toString();
                String columnStr = errorColumn.substring(errorColumn.lastIndexOf("key") + 3, errorColumn.length()).replaceAll("'", "");
                throw ExceptionResultEnum.SQL_ERROR.exception("[" + FieldUniqueEnum.convertToTitle(columnStr) + "]数据不允许重复插入");
            } else if (e instanceof ApiException) {
                ResultUtil.error((ApiException) e, ((ApiException) e).getCode(), e.getMessage());
            } else {
                ResultUtil.error(e.getMessage());
            }
        }
        return ResultUtil.ok(trBasicInfoService.saveOrUpdate(trBasicInfoDb));
    }

    @ApiOperation(value = "导出报告")
    @RequestMapping(value = "/report/export", method = RequestMethod.POST)
    @OperationLogDetail(operationType = OperationTypeEnum.EXPORT)
    @ApiResponses({@ApiResponse(code = 200, message = "下载成功", response = Object.class)})
    @Aac(rateLimit = @RateLimit(count = 1, period = 1000L))
    public void reportExport(@ApiParam(value = "培养方案id", required = true) @RequestParam Long cultureProgramId,
                             @ApiParam(value = "课程id", required = true) @RequestParam Long courseId,
                             @ApiParam(value = "试卷编号") @RequestParam(required = false) String paperNumber) {
        File txtFileTemp = null, fileTemp = null;
        try {
            Long schoolId = Long.valueOf(ServletUtil.getRequestHeaderSchoolId().toString());
            ObeCourseOutline obeCourseOutline = obeCourseOutlineService.findByCultureProgramIdAndCourseId(cultureProgramId, courseId);
            BasicSchool basicSchool = commonCacheService.schoolCache(schoolId);

            TRBasicInfo trBasicInfo = trBasicInfoService.queryBasicInfo(cultureProgramId, courseId, paperNumber);
            Objects.requireNonNull(trBasicInfo, "没有报告信息");
            Objects.requireNonNull(trBasicInfo.getCourseEvaluationResult(), "未设置课程目标，报告生成失败，请在课程大纲管理菜单设置课程目标");
            Objects.requireNonNull(trBasicInfo.getCourseEvaluationResultDetail(), "没有课程考生信息");

            ObeCourseWeightResult obeCourseWeightResult = trBasicInfoService.findCourseWeightResultRmi(obeCourseOutline.getId());
            TCPaperStruct tcPaperStruct = tcPaperStructService.queryPaperStruct(cultureProgramId, courseId, paperNumber);
            if (Objects.nonNull(tcPaperStruct) && Objects.nonNull(tcPaperStruct.getPaperStructDimension())) {
                List<PaperStructDimensionResult> paperStructDimensionResultList = GsonUtil.fromJson(tcPaperStruct.getPaperStructDimension(), new TypeToken<List<PaperStructDimensionResult>>() {
                }.getType());
                for (CourseWeightDto c : obeCourseWeightResult.getSubmitForm()) {
                    Double score = paperStructDimensionResultList.stream().filter(s -> Objects.equals(s.getCourseTargetName(), c.getCourseTargetName())).mapToDouble(s -> s.getScore().doubleValue()).sum();
                    Objects.requireNonNull(c.getTotalWeight(), "[" + c.getCourseTargetName() + "]未设置权重，报告生成失败，请在课程考核设置菜单设置各课程目标所包含考核方式及权重");
                    List<CourseWeightDetailDto> courseWeightDetailDtoList = c.getEvaluationList();
                    for (CourseWeightDetailDto courseWeightDetailDto : courseWeightDetailDtoList) {
                        if (Objects.equals(courseWeightDetailDto.getEvaluationName(), SystemConstant.FINAL_SCORE_STR)) {
                            courseWeightDetailDto.setTargetScore(new BigDecimal(score));
                            break;
                        }
                    }
                }
            }

            ReportCourseEvaluationResultDto reportCourseEvaluationResultDto = JSONObject.parseObject(trBasicInfo.getCourseEvaluationResult(), ReportCourseEvaluationResultDto.class);
            if (Objects.nonNull(reportCourseEvaluationResultDto.getWeightSettingSign()) && reportCourseEvaluationResultDto.getWeightSettingSign().longValue() != obeCourseWeightResult.getWeightSettingSign().longValue()) {
                trBasicInfoService.clearReportData(cultureProgramId, courseId, paperNumber,
                        Objects.nonNull(tcPaperStruct) && Objects.nonNull(tcPaperStruct.getDimensionSign())
                                && tcPaperStruct.getDimensionSign().longValue()
                                != obeCourseWeightResult.getDimensionSign().longValue());
                this.reportView(cultureProgramId, courseId, paperNumber);
                this.reportExport(cultureProgramId, courseId, paperNumber);
            } else {
                //评价样本的基本信息
                JSONObject jsonObject = JSONObject.parseObject(trBasicInfo.getCourseEvaluationResultDetail());
                List<CourseTargetWordDto> courseTargetWordDtoList = JSONArray.parseArray(jsonObject.get("targetWordMap").toString(), CourseTargetWordDto.class);

                StringJoiner courseTarget = new StringJoiner("");
                courseTargetWordDtoList.stream().peek(s -> {
                    courseTarget.add("      ").add(s.getTargetName()).add("——").add(s.getGraduationRequirementPoint()).add("\r\n");
                }).collect(Collectors.toList());

                CourseBasicBean courseBasicBean = new CourseBasicBean(trBasicInfo, courseTarget.toString(), reportCourseEvaluationResultDto.getTargetList().size());

                CourseReportBean courseReportBean = new CourseReportBean(trBasicInfo.getOpenTime() + "《" + trBasicInfo.getCourseName() + "》", courseBasicBean);
                //table1-课程目标目标与毕业要求指标点的对应关系
                courseReportBean.setCourseTargetTable1(trBasicInfoService.buildWordTable1(courseTargetWordDtoList));

                //table2-课程目标达成评价依据
                courseReportBean.setCourseTargetTable2(trBasicInfoService.buildWordTable2(courseTargetWordDtoList));

                //examstudent-课程目标达成评价依据-考生
                courseReportBean.setExamStudentTable1(trBasicInfoService.buildWordTable5(trBasicInfo, courseReportBean));

                InputStream inputStream = null;
                txtFileTemp = SystemConstant.getFileTempVar(SystemConstant.WORD_PREFIX);
                SysConfig sysConfig = commonCacheService.addSysConfigCache(schoolId, SystemConstant.SCHOOL_COURSE_DEGREE_TEMPLATE);
                if (Objects.isNull(sysConfig)) {
                    inputStream = FileUtil.getStream("static/" + ImportTemplateEnum.STATIC_COURSE_DEGREE_REPORT.getTemplateName());
                } else {
                    fileTemp = SystemConstant.getFileTempVar(SystemConstant.WORD_PREFIX);
                    fileTemp = fileUploadService.downloadFile(Long.parseLong(sysConfig.getConfigValue()), fileTemp.getPath());
                    inputStream = new FileInputStream(fileTemp);
                }
                FileUtils.copyInputStreamToFile(inputStream, txtFileTemp);
                XWPFTemplate template = XWPFTemplate.compile(txtFileTemp.getAbsolutePath()).render(courseReportBean);
                template.writeToFile(txtFileTemp.getAbsolutePath());
                // 导出
                String fileName = Objects.nonNull(basicSchool) ? basicSchool.getName() + "_" + courseReportBean.getTitle1() + "_" + ImportTemplateEnum.STATIC_COURSE_DEGREE_REPORT.getFileName() + SystemConstant.WORD_PREFIX : courseReportBean.getTitle1() + "_" + ImportTemplateEnum.STATIC_COURSE_DEGREE_REPORT.getFileName() + SystemConstant.WORD_PREFIX;
                FileUtil.outputFile(ServletUtil.getResponse(), new FileInputStream(txtFileTemp), fileName);
            }
        } catch (Exception e) {
            if (e instanceof ApiException) {
                ResultUtil.error((ApiException) e, ((ApiException) e).getCode(), e.getMessage());
            } else {
                ResultUtil.error(e.getMessage());
            }
        } finally {
            if (Objects.nonNull(txtFileTemp)) {
                txtFileTemp.delete();
            }
            if (Objects.nonNull(fileTemp)) {
                fileTemp.delete();
            }
        }
    }

    @ApiOperation(value = "报告发生改变")
    @RequestMapping(value = "/report/change", method = RequestMethod.POST)
    @ApiResponses({@ApiResponse(code = 200, message = "下载成功", response = Object.class)})
    public Result reportDataChange(@ApiParam(value = "培养方案id", required = true) @RequestParam Long cultureProgramId,
                                   @ApiParam(value = "课程id", required = true) @RequestParam Long courseId,
                                   @ApiParam(value = "试卷编号") @RequestParam(required = false) String paperNumber,
                                   @ApiParam(value = "true：报告，false：成绩管理", required = true) @RequestParam boolean report) {
        ObeCourseOutline obeCourseOutline = obeCourseOutlineService.findByCultureProgramIdAndCourseId(cultureProgramId, courseId);
        ReportChangeResult reportChangeResult = new ReportChangeResult();
        List<TCUsualScore> tcUsualScoreList = tcUsualScoreService.queryUsualScore(cultureProgramId, courseId, paperNumber);
        if (CollectionUtils.isNotEmpty(tcUsualScoreList)) {
            Set<String> usualScoreCourseSet = new LinkedHashSet<>();
            TCUsualScore tcUsualScore = tcUsualScoreList.get(0);
            List<TCUsualScoreDto> tcUsualScoreDtoList = JSONArray.parseArray(tcUsualScore.getScore(), TCUsualScoreDto.class);
            List<String> usualScoreList = tcUsualScoreDtoList.stream().map(s -> s.getName()).collect(Collectors.toList());

            List<CourseWeightDto> courseWeightDtoList = trBasicInfoService.findCourseWeightDtoRmi(obeCourseOutline.getId());
            List<String> courseTargetList = new ArrayList<>(courseWeightDtoList.size());
            if (CollectionUtils.isNotEmpty(courseWeightDtoList)) {
                courseWeightDtoList.stream().peek(e -> {
                    courseTargetList.add(e.getCourseTargetName());
                    List<CourseWeightDetailDto> courseWeightDetailDtoList = e.getEvaluationList();
                    courseWeightDetailDtoList.stream().peek(s -> {
                        if (Objects.nonNull(s.getEnable()) && s.getEnable() && !Objects.equals(s.getEvaluationName(), SystemConstant.FINAL_SCORE_STR)) {
                            usualScoreCourseSet.add(s.getEvaluationName());
                        }
                    }).collect(Collectors.toList());
                }).collect(Collectors.toList());
            }

            if (CollectionUtils.isNotEmpty(usualScoreList) && CollectionUtils.isNotEmpty(usualScoreCourseSet)) {
                List<String> usualScoreCourseList = new ArrayList<>(usualScoreCourseSet);
                Collections.sort(usualScoreList);
                Collections.sort(usualScoreCourseList);
                if (!CollectionUtils.isEqualCollection(usualScoreList, usualScoreCourseList)) {
                    reportChangeResult.setEvaluationChange(true);
                }
            }

            TRBasicInfo trBasicInfo = trBasicInfoService.queryBasicInfo(cultureProgramId, courseId, paperNumber);
            if (Objects.nonNull(trBasicInfo) && Objects.nonNull(trBasicInfo.getCourseEvaluationResultDetail())) {
                JSONObject jsonObject = JSONObject.parseObject(trBasicInfo.getCourseEvaluationResultDetail());
                List<CourseTargetWordDto> courseTargetWordDtoList = JSONArray.parseArray(jsonObject.get("targetWordMap").toString(), CourseTargetWordDto.class);
                List<String> courseTargetDbList = courseTargetWordDtoList.stream().map(
                        CourseTargetWordDto::getTargetName).collect(Collectors.toList());
                if (CollectionUtils.isNotEmpty(courseTargetList) && CollectionUtils.isNotEmpty(courseTargetDbList)) {
                    Collections.sort(courseTargetList);
                    Collections.sort(courseTargetDbList);
                    if (!CollectionUtils.isEqualCollection(courseTargetList, courseTargetDbList)) {
                        reportChangeResult.setCourseTargetChange(true);
                    }
                }
            }

            if (!report) {
                TCPaperStruct tcPaperStruct = tcPaperStructService.queryPaperStruct(cultureProgramId, courseId, paperNumber);
                if (Objects.nonNull(tcPaperStruct) && Objects.nonNull(tcPaperStruct.getPaperStructDimension())) {
//                    List<PaperStructDimensionResult> paperStructDimensionResultList = GsonUtil.fromJson(tcPaperStruct.getPaperStructDimension(), new TypeToken<List<PaperStructDimensionResult>>() {
//                    }.getType());
                    StringJoiner stringJoiner = new StringJoiner("");
                    for (CourseWeightDto c : courseWeightDtoList) {
                        if (Objects.isNull(c.getTotalWeight())) {
                            reportChangeResult.setTargetScoreChange(true);
                            stringJoiner.add("[").add(c.getCourseTargetName()).add("]").add("未设置权重，报告生成失败，请在课程考核设置菜单设置各课程目标所包含考核方式及权重").add("\r\n");
                        }
                    }
                    if (reportChangeResult.isTargetScoreChange()) {
                        trBasicInfoService.clearReportData(cultureProgramId, courseId, paperNumber, false);
                        reportChangeResult.setTargetScoreChangeStr(stringJoiner.toString());
                    }
                }
            }
        }
        return ResultUtil.ok(reportChangeResult);
    }

    @ApiOperation(value = "导出平时成绩报告")
    @RequestMapping(value = "/report/usual_score/export", method = RequestMethod.POST)
    @OperationLogDetail(operationType = OperationTypeEnum.EXPORT)
    @ApiResponses({@ApiResponse(code = 200, message = "下载成功", response = Object.class)})
    @Aac(rateLimit = @RateLimit(count = 1, period = 1000L))
    public void reportUsualscoreExport(@ApiParam(value = "培养方案id", required = true) @RequestParam Long cultureProgramId,
                                       @ApiParam(value = "课程id", required = true) @RequestParam Long courseId,
                                       @ApiParam(value = "试卷编号") @RequestParam(required = false) String paperNumber) {
        TRBasicInfo trBasicInfo = trBasicInfoService.queryBasicInfo(cultureProgramId, courseId, paperNumber);
        Objects.requireNonNull(trBasicInfo, "没有报告信息");
        Objects.requireNonNull(trBasicInfo.getCourseEvaluationResult(), "没有课程目标信息");
        Objects.requireNonNull(trBasicInfo.getCourseEvaluationResultDetail(), "没有课程考生信息");

        List<TRExamStudent> trExamStudentList = trExamStudentService.list(new QueryWrapper<TRExamStudent>().lambda()
                .eq(TRExamStudent::getrBasicInfoId, trBasicInfo.getId()).and(w -> w.notIn(TRExamStudent::getStudentCode, "各课程目标平均分", "平均分", "目标分")));
        if (CollectionUtils.isEmpty(trExamStudentList)) {
            throw ExceptionResultEnum.ERROR.exception("未找到学生信息");
        }

        Map<Long, BigDecimal> examStudentMatrixDegree = new LinkedHashMap<>(), examStudentSumMatrixDegree = new LinkedHashMap<>();
        BigDecimal examStudentSumTargetWeight = new BigDecimal(0);
        LinkedMultiValueMap<String, String> courseTargetMap = new LinkedMultiValueMap<>();
        LinkedMultiValueMap<Long, BigDecimal> examStudentMatrixDegreeMap = new LinkedMultiValueMap<>();
        for (int y = 0; y < trExamStudentList.size(); y++) {
            TRExamStudent t = trExamStudentList.get(y);
            if (Objects.nonNull(t.getResultDetail())) {
                List<ReportExamStudentTargetDto> reportExamStudentTargetDtoList = JSONArray.parseArray(t.getResultDetail(), ReportExamStudentTargetDto.class);
                for (ReportExamStudentTargetDto r : reportExamStudentTargetDtoList) {
                    if (Objects.nonNull(r.getUsualScore()) && CollectionUtils.isNotEmpty(r.getUsualScore().getScoreList())) {
                        for (int i = 0; i < r.getUsualScore().getScoreList().size(); i++) {
                            ReportExamStudentUsualScoreObjDto reportExamStudentUsualScoreObjDto = r.getUsualScore().getScoreList().get(i);
                            //课程目标有几个
                            if (!courseTargetMap.containsKey(r.getTargetName())) {
                                courseTargetMap.add(r.getTargetName(), reportExamStudentUsualScoreObjDto.getEvaluation());
                            } else {
                                List<String> list = courseTargetMap.get(r.getTargetName());
                                if (!list.contains(reportExamStudentUsualScoreObjDto.getEvaluation())) {
                                    courseTargetMap.add(r.getTargetName(), reportExamStudentUsualScoreObjDto.getEvaluation());
                                }
                            }
                            //课程目标有几个
                            examStudentMatrixDegreeMap.add(t.getId(), reportExamStudentUsualScoreObjDto.getMatrixScore());
                            //各个学生的换算分数
                            if (!examStudentMatrixDegree.containsKey(t.getId())) {
                                examStudentMatrixDegree.put(t.getId(), reportExamStudentUsualScoreObjDto.getMatrixScore());
                            } else {
                                BigDecimal matrixDegree = examStudentMatrixDegree.get(t.getId());
                                matrixDegree = matrixDegree.add(reportExamStudentUsualScoreObjDto.getMatrixScore());
                                examStudentMatrixDegree.put(t.getId(), matrixDegree);
                            }
                            //权重
                            if (y == 0) {
                                examStudentSumTargetWeight = examStudentSumTargetWeight.add(reportExamStudentUsualScoreObjDto.getTargetWeight());
                            }
                            if (i == r.getUsualScore().getScoreList().size() - 1) {
                                BigDecimal matrixDegree = examStudentMatrixDegree.get(t.getId());
                                matrixDegree = matrixDegree.setScale(2, BigDecimal.ROUND_HALF_UP);
                                examStudentMatrixDegree.put(t.getId(), matrixDegree);

                                examStudentSumTargetWeight = examStudentSumTargetWeight.setScale(2, BigDecimal.ROUND_HALF_UP);
                                BigDecimal sumMatrixDegree = SystemConstant.PERCENT.multiply(matrixDegree).divide(examStudentSumTargetWeight, 2, BigDecimal.ROUND_HALF_UP);
                                examStudentSumMatrixDegree.put(t.getId(), sumMatrixDegree);
                            }
                        }
                    }
                }
            }
        }

        SXSSFWorkbook wb = new SXSSFWorkbook();
        Sheet sheet = wb.createSheet("过程考核");
        CellStyle headerStyle = this.createHeadStyle(wb);

        Row rowHeadOne = sheet.createRow(0);
        Row rowHeadTwo = sheet.createRow(1);
        Row rowOne = sheet.createRow(2);
        Row rowTwo = sheet.createRow(3);

        Long schoolId = Long.valueOf(ServletUtil.getRequestHeaderSchoolId().toString());
        BasicSchool basicSchool = commonCacheService.schoolCache(schoolId);
        String fileName = Objects.nonNull(basicSchool) ? basicSchool.getName() + "_" + trBasicInfo.getOpenTime() + "《" + trBasicInfo.getCourseName() + "》_平时成绩" : trBasicInfo.getOpenTime() + "《" + trBasicInfo.getCourseName() + "》_平时成绩";

        //姓名
        this.drawExcelFixedHead(rowOne, headerStyle, sheet, "姓名", 0, 2, 3, 0, 0, sheet.getColumnWidth(0) * 17 / 10);

        AtomicInteger cellOneIndex = new AtomicInteger(1);
        AtomicInteger oneFirst = new AtomicInteger(1);
        AtomicInteger oneLast = new AtomicInteger(0);
        AtomicInteger cellTwoIndex = new AtomicInteger(1);
        for (Map.Entry<String, List<String>> entry : courseTargetMap.entrySet()) {
            List<String> childKeys = entry.getValue();
            String key = entry.getKey();
            Cell cell = rowOne.createCell(cellOneIndex.get());
            cell.setCellValue(key);
            cell.setCellStyle(headerStyle);
            oneLast.compareAndSet(oneLast.get(), oneLast.get() + childKeys.size());
            if (oneFirst.get() != oneLast.get()) {
                CellRangeAddress region = new CellRangeAddress(2, 2, (short) oneFirst.get(), (short) oneLast.get());
                this.setRegionBorder(region, sheet);
                sheet.addMergedRegion(region);
            }
            sheet.setColumnWidth(cellOneIndex.get(), sheet.getColumnWidth(cellOneIndex.get()) * 17 / 10);
            cellOneIndex.compareAndSet(cellOneIndex.get(), cellOneIndex.get() + childKeys.size());
            oneFirst.compareAndSet(oneFirst.get(), cellOneIndex.get());
            for (String s : childKeys) {
                this.drawExcelFixedHead(rowTwo, headerStyle, sheet, s, cellTwoIndex.get(), sheet.getColumnWidth(cellTwoIndex.get()) * 17 / 10);
                cellTwoIndex.incrementAndGet();
            }
        }

        this.drawExcelFixedHead(rowOne, headerStyle, sheet, "平时成绩总分(满分" + SystemConstant.df.format(examStudentSumTargetWeight) + "分）", cellTwoIndex.get(), 2, 3, cellTwoIndex.get(), cellTwoIndex.get(), sheet.getColumnWidth(cellTwoIndex.get()) * 17 / 5);
        cellTwoIndex.incrementAndGet();
        this.drawExcelFixedHead(rowOne, headerStyle, sheet, "平时成绩总分(满分100分）", cellTwoIndex.get(), 2, 3, cellTwoIndex.get(), cellTwoIndex.get(), sheet.getColumnWidth(cellTwoIndex.get()) * 17 / 5);
        cellTwoIndex.incrementAndGet();

        //第一行
        this.drawExcelFixedHead(rowHeadOne, headerStyle, sheet, "《" + trBasicInfo.getCourseName() + "》课程过程考核原始记录", 0, 0, 0, 0, cellTwoIndex.get() - 1, sheet.getColumnWidth(cellTwoIndex.get()) * 17);
        //第二行
        this.drawExcelFixedHead(rowHeadTwo, headerStyle, sheet, "课程：" + trBasicInfo.getCourseName() + "     班级：" + (Objects.nonNull(trBasicInfo.getTeachingObject()) ? trBasicInfo.getTeachingObject() : "无") + "     任课教师：" + (Objects.nonNull(trBasicInfo.getTeacher()) ? trBasicInfo.getTeacher() : "无") + "     学期：" + trBasicInfo.getOpenTime(), 0, 1, 1, 0, cellTwoIndex.get() - 1, sheet.getColumnWidth(cellTwoIndex.get()) * 17);

        CellStyle styleExamStudent = this.createRowStyle(wb);
        AtomicInteger rowStudentIndex = new AtomicInteger(4);
        for (TRExamStudent t : trExamStudentList) {
            Row rowStudent = sheet.createRow(rowStudentIndex.get());
            this.drawExcelFixedHead(rowStudent, styleExamStudent, sheet, t.getName(), 0, sheet.getColumnWidth(2) * 17 / 10);

            List<BigDecimal> bigDecimalList = examStudentMatrixDegreeMap.get(t.getId());
            if (CollectionUtils.isNotEmpty(bigDecimalList)) {
                for (int i = 0; i < bigDecimalList.size(); i++) {
                    this.drawExcelFixedHead(rowStudent, styleExamStudent, sheet, SystemConstant.df.format(bigDecimalList.get(i)), i + 1, sheet.getColumnWidth(i + 1) * 17 / 10);
                }
                this.drawExcelFixedHead(rowStudent, styleExamStudent, sheet, SystemConstant.df.format(!org.springframework.util.CollectionUtils.isEmpty(examStudentMatrixDegree) && examStudentMatrixDegree.containsKey(t.getId()) ? examStudentMatrixDegree.get(t.getId()) : 0), bigDecimalList.size() + 1, sheet.getColumnWidth(bigDecimalList.size() + 1) * 17 / 10);
                this.drawExcelFixedHead(rowStudent, styleExamStudent, sheet, SystemConstant.df.format(!org.springframework.util.CollectionUtils.isEmpty(examStudentSumMatrixDegree) && examStudentSumMatrixDegree.containsKey(t.getId()) ? examStudentSumMatrixDegree.get(t.getId()) : 0), bigDecimalList.size() + 2, sheet.getColumnWidth(bigDecimalList.size() + 2) * 17 / 10);
            }
            rowStudentIndex.incrementAndGet();
        }
        ExcelUtil.exportEXCEL(fileName, wb, ServletUtil.getResponse());
    }

    /**
     * 创建行样式
     *
     * @param wb
     * @return
     */
    protected CellStyle createRowStyle(SXSSFWorkbook wb) {
        CellStyle styleExamStudent = wb.createCellStyle();
        styleExamStudent.setAlignment(HorizontalAlignment.LEFT);
        styleExamStudent.setVerticalAlignment(VerticalAlignment.CENTER);
        return styleExamStudent;
    }

    /**
     * 创建头样式
     *
     * @param wb
     * @return
     */
    protected CellStyle createHeadStyle(SXSSFWorkbook wb) {
        CellStyle headerStyle = wb.createCellStyle();
        headerStyle.setAlignment(HorizontalAlignment.CENTER); // 水平居中格式
        headerStyle.setVerticalAlignment(VerticalAlignment.CENTER); //垂直居中
        headerStyle.setBorderRight(BorderStyle.THIN);
        headerStyle.setRightBorderColor(IndexedColors.BLACK.getIndex());
        headerStyle.setBorderLeft(BorderStyle.THIN);
        headerStyle.setLeftBorderColor(IndexedColors.BLACK.getIndex());
        headerStyle.setBorderTop(BorderStyle.THIN);
        headerStyle.setTopBorderColor(IndexedColors.BLACK.getIndex());
        headerStyle.setBorderBottom(BorderStyle.THIN);
        headerStyle.setBottomBorderColor(IndexedColors.BLACK.getIndex());
        // 背景颜色
        headerStyle.setFillPattern(FillPatternType.SOLID_FOREGROUND);
        headerStyle.setFillForegroundColor(IndexedColors.GREY_25_PERCENT.getIndex());

        Font defaultFont = wb.createFont();
        defaultFont.setFontHeightInPoints((short) 12);
        defaultFont.setBold(true);
        headerStyle.setFont(defaultFont);
        return headerStyle;
    }

    /**
     * 创建固定行表头
     *
     * @param row
     * @param style
     * @param sheet
     * @param cellName
     * @param cellIndex
     * @param firstRow
     * @param lastRow
     * @param firstCol
     * @param lastCol
     * @param columnWidth
     */
    protected void drawExcelFixedHead(Row row, CellStyle style, Sheet sheet, String cellName, int cellIndex,
                                      int firstRow, int lastRow, int firstCol, int lastCol, int columnWidth) {
        Cell cell = row.createCell(cellIndex);
        cell.setCellValue(cellName);
        cell.setCellStyle(style);
        CellRangeAddress region = new CellRangeAddress(firstRow, lastRow, (short) firstCol, (short) lastCol);
        sheet.addMergedRegion(region);
        this.setRegionBorder(region, sheet);
        sheet.setColumnWidth(cellIndex, columnWidth);
    }

    /**
     * 设置跨行边框
     *
     * @param region
     * @param sheet
     */
    protected void setRegionBorder(CellRangeAddress region, Sheet sheet) {
        RegionUtil.setBorderLeft(BorderStyle.THIN, region, sheet);//设置左边框
        RegionUtil.setBorderBottom(BorderStyle.THIN, region, sheet);//设置下边框
        RegionUtil.setBorderRight(BorderStyle.THIN, region, sheet);//设置右边框
        RegionUtil.setBorderTop(BorderStyle.THIN, region, sheet);//设置左边框
    }

    /**
     * 创建动态固定行表头
     *
     * @param row
     * @param style
     * @param sheet
     * @param cellName
     * @param cellIndex
     * @param columnWidth
     */
    protected void drawExcelFixedHead(Row row, CellStyle style, Sheet sheet, String cellName, int cellIndex, int columnWidth) {
        Cell cell = row.createCell(cellIndex);
        cell.setCellValue(cellName);
        cell.setCellStyle(style);
        sheet.setColumnWidth(cellIndex, columnWidth > 10000 ? (int) Math.abs(columnWidth / 2) : columnWidth);
    }

    @Resource
    WordToPdfUtil wordToPdfUtil;

    @Resource
    CustomRedisLockProvider customRedisLockProvider;

    @ApiOperation(value = "wordtopdf测试")
    @RequestMapping(value = "/wordTopdf", method = RequestMethod.POST)
    @ApiResponses({@ApiResponse(code = 200, message = "转换成功", response = Object.class)})
    @Aac(auth = false)
    public Result wordTopdfTest() throws Exception {
        if (customRedisLockProvider.tryLock(LockType.CREATE_PDF, "1")) {
            try {
                if (!customRedisLockProvider.isWriteLocked(LockType.CREATE_PDF, "1")) {
                    customRedisLockProvider.waitLock(LockType.CREATE_PDF, "1");
                }
                customRedisLockProvider.watch(LockType.CREATE_PDF, "1");
                customRedisLockProvider.watch(LockType.CREATE_PDF, "1");
                wordToPdfUtil.wordToPdf("/Users/king/Downloads/测试学校_1_2021~2022第一学期_上学期《电力电子与电力传动》_课程目标达成度.docx", "/Users/king/Downloads/测试学校_1_2021~2022第一学期_上学期《电力电子与电力传动》_课程目标达成度.pdf", 0);
                File file = new File("/Users/king/Downloads/测试学校_1_2021~2022第一学期_上学期《电力电子与电力传动》_课程目标达成度.pdf");
//                PdfUtil.addWaterMarkFullScreen(file, new String[]{"启明泰和_课程目标达成度测试水印"});
                PdfUtil.waterMark(file, "启明泰和_课程目标达成度测试水印", "123");
            } finally {
                customRedisLockProvider.unlock(LockType.CREATE_PDF, "1");
                customRedisLockProvider.unwatch(LockType.CREATE_PDF, "1");
                customRedisLockProvider.unwatch(LockType.CREATE_PDF, "1");
            }
        }
        return ResultUtil.ok(true);
    }

    public static boolean threadBreak = false;
    public static CopyOnWriteArrayList<Integer> threadKeyList = new CopyOnWriteArrayList<>();
    public static String path = "/Users/king/Downloads/";
    public static File fileDir = new File(path + "thread_test");
    public static final String lockName = ":txt:file";
    public static final String one = "[一]";
    public static final String two = "[二]";
    public static ConcurrentHashMap<Integer, Integer> keyMap = new ConcurrentHashMap<>();

    @ApiOperation(value = "线程测试")
    @RequestMapping(value = "/thread/test", method = RequestMethod.POST)
    @ApiResponses({@ApiResponse(code = 200, message = "测试成功", response = Object.class)})
    @Aac(auth = false)
    public Result threadTest() throws IOException, Exception {
        threadBreak = false;
        AtomicInteger threadReleaseToatlCount = new AtomicInteger(0);
        threadKeyList.clear();
        keyMap.clear();
        int num = 100;
        for (; ; ) {
            int key = new Random().nextInt(num) + 1;
            if (!keyMap.containsKey(key)) {
                keyMap.put(key, key);
                new Thread(() -> {
                    //开始尝试获取锁
                    if (customRedisLockProvider.tryLock(LockType.CREATE_PDF, key + "", 1L, TimeUnit.DAYS)) {
                        log.info("########线程key:{}已获得锁########", key);
                        try {
                            StringJoiner stringJoinerSummary = this.readTxtFile(key + "");
                            this.writeTxtFile(stringJoinerSummary, key + "", one + "->线程key:" + key + "已获得锁");
                        } catch (IOException | InterruptedException e) {
                            e.printStackTrace();
                        }
                        try {
                            threadKeyList.add(key);
                            long i = 0;
                            do {
                                i = i + new Random().nextInt(100) + 1;
                                if ((Thread.currentThread().getId() & 1) == 0) {
                                    Thread.sleep(new Random().nextInt(100) + 1);
                                } else {
                                    Thread.sleep(new Random().nextInt(50) + 1);
                                }
                            } while (i < 100000);
                        } catch (InterruptedException e) {
                            e.printStackTrace();
                        } finally {
                            customRedisLockProvider.unlock(LockType.CREATE_PDF, key + "");
                            threadReleaseToatlCount.incrementAndGet();
                            log.info("########线程key:{}已被释放,总计:{}########", key, threadReleaseToatlCount.get());
                            try {
                                StringJoiner stringJoinerSummary = this.readTxtFile(key + "");
                                this.writeTxtFile(stringJoinerSummary, key + "", one + "->线程key:" + key + "已被释放");
                            } catch (InterruptedException | IOException e) {
                                e.printStackTrace();
                            }
                            if (threadReleaseToatlCount.get() == num) {
                                threadBreak = true;
                            }
                        }
                    }
                }).start();
            }
            Thread.sleep(1000);
            if (keyMap.size() == num) {
                log.info("########线程生成完毕########");
                break;
            }
        }
        return ResultUtil.ok(true);
    }

    @ApiOperation(value = "线程抢夺测试")
    @RequestMapping(value = "/thread/reave/test", method = RequestMethod.POST)
    @ApiResponses({@ApiResponse(code = 200, message = "测试成功", response = Object.class)})
    @Aac(auth = false)
    public Result threadReaveTest() throws IOException, Exception {
        AtomicInteger threadReaveToatlCount = new AtomicInteger(0);
        //开始尝试获取锁
        for (; ; ) {
            int index = new Random().nextInt(threadKeyList.size());
            int key = threadKeyList.get(index);
            if (customRedisLockProvider.tryLock(LockType.CREATE_PDF, key + "")) {
                log.info("%%%%%%%%%线程key:{}抢到锁%%%%%%%%%", key);
                StringJoiner stringJoinerSummary = this.readTxtFile(key + "");
                this.writeTxtFile(stringJoinerSummary, key + "", two + "->线程key:" + key + "抢到锁");
                threadReaveToatlCount.incrementAndGet();
                threadKeyList.remove(key);
                customRedisLockProvider.unlock(LockType.CREATE_PDF, key + "");
            } else {
                log.info("%%%%%%%%%线程key:{}正被锁定%%%%%%%%%", key);
                StringJoiner stringJoinerSummary = this.readTxtFile(key + "");
                this.writeTxtFile(stringJoinerSummary, key + "", two + "->线程key:" + key + "正被锁定");
            }
            Thread.sleep(1000);
            if (threadKeyList.size() == 0 || threadBreak) {
                log.info("%%%%%%%%%线程抢夺完毕%%%%%%%%%,总计:{}", threadReaveToatlCount.get());
                break;
            }
        }
        return ResultUtil.ok(true);
    }

    /**
     * 获取txt文件读锁
     *
     * @param key
     * @return
     * @throws IOException
     */
    protected StringJoiner readTxtFile(String key) throws IOException {
        if (!fileDir.exists()) {
            fileDir.mkdirs();
        }
        File file = new File(fileDir.getPath() + File.separator + key + ".txt");
        if (!file.exists()) {
            file.createNewFile();
        }
        StringJoiner stringJoinerSummary = new StringJoiner("\n");
        try {
            customRedisLockProvider.watch(LockType.CREATE_PDF, key + lockName + ":read");
            String text = IOUtils.toString(new FileInputStream(file), Charset.forName(SystemConstant.CHARSET_NAME));
            if (Objects.nonNull(text) && !Objects.equals(text.trim(), "")) {
                stringJoinerSummary.add(text);
            }
        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            customRedisLockProvider.unwatch(LockType.CREATE_PDF, key + lockName + ":read");
        }
        return stringJoinerSummary;
    }

    /**
     * txt文件写锁
     *
     * @param stringJoinerSummary
     * @param key
     * @param message
     * @throws InterruptedException
     */
    protected void writeTxtFile(StringJoiner stringJoinerSummary, String key, String message) throws InterruptedException, IOException {
        if (!fileDir.exists()) {
            fileDir.mkdirs();
        }
        File file = new File(fileDir.getPath() + File.separator + key + ".txt");
        if (!file.exists()) {
            file.createNewFile();
        }
        for (; ; ) {
            if (!customRedisLockProvider.isWriteLocked(LockType.CREATE_PDF, key + lockName + ":wirte")) {
                try {
                    customRedisLockProvider.waitLock(LockType.CREATE_PDF, key + lockName + ":wirte");
                    stringJoinerSummary.add(MessageFormat.format("{0}{1}", DateUtil.format(new Date(), SystemConstant.DEFAULT_DATE_PATTERN), message));
                    IOUtils.write(stringJoinerSummary.toString().getBytes(StandardCharsets.UTF_8), new FileOutputStream(file));
                } catch (IOException e) {
                    e.printStackTrace();
                } finally {
                    customRedisLockProvider.unlock(LockType.CREATE_PDF, key + lockName + ":wirte");
                    break;
                }
            }
            Thread.sleep(500);
        }
    }
}
